"""Defines the Data Classes used."""
from datetime import datetime, timezone
import math

from pyastroweatherio.const import (
    CLOUDCOVER_PLAIN,
    CONDITION,
    CONDITION_PLAIN,
    DEEP_SKY_THRESHOLD,
    LIFTED_INDEX_PLAIN,
    SEEING_PLAIN,
    TRANSPARENCY_PLAIN,
    WIND10M_DIRECTON,
)


class BaseData:
    """A representation of the base class for AstroWeather Data."""

    def __init__(self, data):
        self._init = data["init"]
        self._timepoint = data["timepoint"]
        self._timestamp = data["timestamp"]
        self._cloudcover = data["cloudcover"]
        self._cloud_area_fraction = data["cloud_area_fraction"]
        self._cloud_area_fraction_high = data["cloud_area_fraction_high"]
        self._cloud_area_fraction_low = data["cloud_area_fraction_low"]
        self._cloud_area_fraction_medium = data["cloud_area_fraction_medium"]
        self._fog_area_fraction = data["fog_area_fraction"]
        self._seeing = data["seeing"]
        self._transparency = data["transparency"]
        self._condition_percentage = data["condition_percentage"]
        self._lifted_index = data["lifted_index"]
        self._rh2m = data["rh2m"]
        self._wind_speed = data["wind_speed"]
        self._wind_from_direction = data["wind_from_direction"]
        self._temp2m = data["temp2m"]
        self._dewpoint2m = data["dewpoint2m"]
        self._weather = data["weather"]
        self._weather6 = data["weather6"]
        self._precipitation_amount = data["precipitation_amount"]

    @property
    def init(self) -> datetime:
        """Return Forecast Anchor."""
        return self._init.replace(microsecond=0, tzinfo=timezone.utc)

    @property
    def timepoint(self) -> int:
        """Return Forecast Hour."""
        return self._timepoint

    @property
    def timestamp(self) -> datetime:
        """Return Forecast Timestamp."""
        return self._timestamp.replace(microsecond=0, tzinfo=timezone.utc)

    @property
    def condition_percentage(self) -> int:
        """Return condition based on cloud cover, seeing and transparency"""
        return self._condition_percentage

    @property
    def cloudcover(self) -> int:
        """Return Cloud Coverage."""
        return self._cloudcover

    @property
    def cloudcover_percentage(self) -> int:
        """Return Cloud Cover Percentage."""
        return int(-(100 * (1 - self._cloudcover) / 8))

    @property
    def cloudless_percentage(self) -> int:
        """Return Cloudless Percentage."""
        return int(100 + 100 * (1 - self._cloudcover) / 8)

    @property
    def cloud_area_fraction_percentage(self) -> int:
        """Return Cloud Cover Percentage."""
        return self._cloud_area_fraction

    @property
    def cloud_area_fraction_high_percentage(self) -> int:
        """Return Cloud Cover Percentage."""
        return self._cloud_area_fraction_high

    @property
    def cloud_area_fraction_low_percentage(self) -> int:
        """Return Cloud Cover Percentage."""
        return self._cloud_area_fraction_low

    @property
    def cloud_area_fraction_medium_percentage(self) -> int:
        """Return Cloud Cover Percentage."""
        return self._cloud_area_fraction_medium

    @property
    def fog_area_fraction_percentage(self) -> int:
        """Return Fog Area Percentage."""
        return self._fog_area_fraction

    @property
    def seeing(self) -> int:
        """Return Seeing."""
        return self._seeing

    @property
    def seeing_percentage(self) -> int:
        """Return Seeing."""
        return int((100 + 100 / 7 - self._seeing * 100 / 7))

    @property
    def transparency(self) -> int:
        """Return Transparency."""
        return self._transparency

    @property
    def transparency_percentage(self) -> int:
        """Return Transparency."""
        return int((100 + 100 / 7 - self._transparency * 100 / 7))

    @property
    def lifted_index(self) -> int:
        """Return Lifted Index."""
        return self._lifted_index

    @property
    def rh2m(self) -> int:
        """Return 2m Relative Humidity."""
        return self._rh2m

    @property
    def wind10m_speed(self) -> float:
        """Return 10m Wind Speed."""
        return self._wind_speed

    @property
    def wind10m_direction(self) -> str:
        """Return 10m Wind Direction."""
        direction = self._wind_from_direction
        direction += 22.5
        direction = direction % 360
        direction = int(direction / 45)  # values 0 to 7
        return WIND10M_DIRECTON[direction]

    @property
    def temp2m(self) -> int:
        """Return 2m Temperature."""
        return self._temp2m

    @property
    def dewpoint2m(self) -> float:
        """Return 2m Dew Point."""
        return round(self._dewpoint2m, 1)

    @property
    def weather(self) -> str:
        """Return Current Weather."""
        return self._weather.replace("_", " ").capitalize()

    @property
    def weather6(self) -> str:
        """Return Current Weather."""
        return self._weather6.replace("_", " ").capitalize()

    @property
    def precipitation_amount(self) -> float:
        """Return Current Weather."""
        return self._precipitation_amount


class LocationData(BaseData):
    """A representation of the Location AstroWeather Data."""

    def __init__(self, data):
        super().__init__(data)
        self._forecast_length = data["forecast_length"]
        self._latitude = data["latitude"]
        self._longitude = data["longitude"]
        self._elevation = data["elevation"]
        self._sun_next_rising = data["sun_next_rising"]
        self._sun_next_rising_nautical = data["sun_next_rising_nautical"]
        self._sun_next_rising_astro = data["sun_next_rising_astro"]
        self._sun_next_setting = data["sun_next_setting"]
        self._sun_next_setting_nautical = data["sun_next_setting_nautical"]
        self._sun_next_setting_astro = data["sun_next_setting_astro"]
        self._sun_altitude = data["sun_altitude"]
        self._sun_azimuth = data["sun_azimuth"]
        self._moon_next_rising = data["moon_next_rising"]
        self._moon_next_setting = data["moon_next_setting"]
        self._moon_phase = data["moon_phase"]
        self._moon_next_new_moon = data["moon_next_new_moon"]
        self._moon_next_full_moon = data["moon_next_full_moon"]
        self._moon_altitude = data["moon_altitude"]
        self._moon_azimuth = data["moon_azimuth"]
        self._night_duration_astronomical = data["night_duration_astronomical"]
        self._deep_sky_darkness_moon_rises = data["deep_sky_darkness_moon_rises"]
        self._deep_sky_darkness_moon_sets = data["deep_sky_darkness_moon_sets"]
        self._deep_sky_darkness_moon_always_up = data["deep_sky_darkness_moon_always_up"]
        self._deep_sky_darkness_moon_always_down = data["deep_sky_darkness_moon_always_down"]
        self._deep_sky_darkness = data["deep_sky_darkness"]
        self._deepsky_forecast = data["deepsky_forecast"]
        self._uptonight = data["uptonight"]

    @property
    def forecast_length(self) -> int:
        """Return Forecast Length in Hours"""
        return self._forecast_length

    @property
    def latitude(self) -> float:
        """Return Latitude."""
        return self._latitude

    @property
    def longitude(self) -> float:
        """Return Longitude."""
        return self._longitude

    @property
    def elevation(self) -> float:
        """Return Elevation."""
        return self._elevation

    @property
    def cloudcover_plain(self) -> str:
        """Return Cloud Coverage."""
        cover = self._cloudcover
        if cover >= 1 and cover <= 9:
            return CLOUDCOVER_PLAIN[cover - 1]
        return None

    @property
    def seeing_plain(self) -> str:
        """Return Seeing."""
        seeing = self._seeing
        if seeing >= 1 and seeing <= 8:
            return SEEING_PLAIN[seeing - 1]
        return None

    @property
    def transparency_plain(self) -> str:
        """Return Transparency."""
        transparency = self._transparency
        if transparency >= 1 and transparency <= 8:
            return TRANSPARENCY_PLAIN[self._transparency - 1]
        return None

    @property
    def lifted_index_plain(self) -> str:
        """Return Lifted Index."""

        trans = {
            -10: LIFTED_INDEX_PLAIN[0],
            -6: LIFTED_INDEX_PLAIN[1],
            -4: LIFTED_INDEX_PLAIN[2],
            -1: LIFTED_INDEX_PLAIN[3],
            2: LIFTED_INDEX_PLAIN[4],
            6: LIFTED_INDEX_PLAIN[5],
            10: LIFTED_INDEX_PLAIN[6],
            15: LIFTED_INDEX_PLAIN[7],
        }
        lifted_index = self._lifted_index
        if lifted_index in trans:
            return trans.get(self._lifted_index, "")
        return None

    @property
    def deep_sky_view(self) -> bool:
        """Return True if Deep Sky should be possible."""
        if self.condition_percentage >= DEEP_SKY_THRESHOLD:
            return True
        return False

    @property
    def condition_plain(self) -> str:
        """Return Current View Conditions."""
        if self.condition_percentage > 80:
            return CONDITION_PLAIN[0].capitalize()
        if self.condition_percentage > 60:
            return CONDITION_PLAIN[1].capitalize()
        if self.condition_percentage > 40:
            return CONDITION_PLAIN[2].capitalize()
        if self.condition_percentage > 20:
            return CONDITION_PLAIN[3].capitalize()
        return CONDITION_PLAIN[4].capitalize()

    @property
    def sun_next_rising(self) -> datetime:
        """Return Sun Next Rising Civil."""
        if isinstance(self._sun_next_rising, datetime):
            return self._sun_next_rising

    @property
    def sun_next_rising_nautical(self) -> datetime:
        """Return Sun Next Rising Nautical."""
        if isinstance(self._sun_next_rising_nautical, datetime):
            return self._sun_next_rising_nautical

    @property
    def sun_next_rising_astro(self) -> datetime:
        """Return Sun Next Rising Astronomical."""
        if isinstance(self._sun_next_rising_astro, datetime):
            return self._sun_next_rising_astro

    @property
    def sun_next_setting(self) -> datetime:
        """Return Next Setting Civil."""
        if isinstance(self._sun_next_setting, datetime):
            return self._sun_next_setting

    @property
    def sun_next_setting_nautical(self) -> datetime:
        """Return Sun Next Setting Nautical."""
        if isinstance(self._sun_next_setting_nautical, datetime):
            return self._sun_next_setting_nautical

    @property
    def sun_next_setting_astro(self) -> datetime:
        """Return Sun Next Setting Astronomical."""
        if isinstance(self._sun_next_setting_astro, datetime):
            return self._sun_next_setting_astro

    @property
    def sun_altitude(self) -> float:
        """Return Sun Altitude."""
        return round(self._sun_altitude, 3)

    @property
    def sun_azimuth(self) -> float:
        """Return sun Azimuth."""
        return round(self._sun_azimuth, 3)

    @property
    def moon_next_rising(self) -> datetime:
        """Return Moon Next Rising."""
        if isinstance(self._moon_next_rising, datetime):
            return self._moon_next_rising

    @property
    def moon_next_setting(self) -> datetime:
        """Return Moon Next Setting."""
        if isinstance(self._moon_next_setting, datetime):
            return self._moon_next_setting

    @property
    def moon_phase(self) -> float:
        """Return Moon Phase."""
        return round(self._moon_phase, 1)

    @property
    def moon_next_new_moon(self) -> datetime:
        """Return Moon Next New Moon."""
        if isinstance(self._moon_next_new_moon, datetime):
            return self._moon_next_new_moon

    @property
    def moon_next_full_moon(self) -> datetime:
        """Return Moon Next Full Moon."""
        if isinstance(self._moon_next_full_moon, datetime):
            return self._moon_next_full_moon
        
    @property
    def moon_altitude(self) -> float:
        """Return Moon Altitude."""
        return round(self._moon_altitude, 3)

    @property
    def moon_azimuth(self) -> float:
        """Return Moon Azimuth."""
        return round(self._moon_azimuth, 3)

    @property
    def night_duration_astronomical(self) -> float:
        """Returns the remaining timespan of astronomical darkness."""
        return self._night_duration_astronomical

    @property
    def deep_sky_darkness_moon_rises(self) -> bool:
        """Returns true if moon rises during astronomical night."""
        return self._deep_sky_darkness_moon_rises

    @property
    def deep_sky_darkness_moon_sets(self) -> bool:
        """Returns true if moon sets during astronomical night."""
        return self._deep_sky_darkness_moon_sets

    @property
    def deep_sky_darkness_moon_always_up(self) -> bool:
        """Returns true if moon is up during astronomical night."""
        return self._deep_sky_darkness_moon_always_up

    @property
    def deep_sky_darkness_moon_always_down(self) -> bool:
        """Returns true if moon is down during astronomical night."""
        return self._deep_sky_darkness_moon_always_down

    @property
    def deep_sky_darkness(self) -> float:
        """Returns the remaining timespan of deep sky darkness."""
        return self._deep_sky_darkness

    @property
    def deepsky_forecast_today(self) -> int:
        """Return Forecas Today in Percent."""
        nightly_condition_sum = 0
        for nightly_condition in self._deepsky_forecast[0].nightly_conditions:
            nightly_condition_sum += nightly_condition
        return int(round(nightly_condition_sum / len(self._deepsky_forecast[0].nightly_conditions)))

    @property
    def deepsky_forecast_today_dayname(self):
        """Return Forecast Todays Dayname."""
        if len(self._deepsky_forecast) > 0:
            nightly_conditions = self._deepsky_forecast[0]
            return nightly_conditions.dayname
        return None

    @property
    def deepsky_forecast_today_plain(self):
        """Return Forecast Today."""
        out = ""
        for nightly_condition in self._deepsky_forecast[0].nightly_conditions:
            out += CONDITION[4 - math.floor(nightly_condition / 20)].capitalize()
        return out

    @property
    def deepsky_forecast_today_desc(self):
        """Return Forecast Today Description."""
        if len(self._deepsky_forecast) > 0:
            nightly_conditions = self._deepsky_forecast[0]
            return nightly_conditions.weather.replace("_", " ").capitalize()
        return None

    @property
    def deepsky_forecast_tomorrow(self) -> int:
        """Return Forecas Tomorrow in Percentt."""
        nightly_condition_sum = 0
        for nightly_condition in self._deepsky_forecast[1].nightly_conditions:
            nightly_condition_sum += nightly_condition
        return int(round(nightly_condition_sum / len(self._deepsky_forecast[1].nightly_conditions)))

    @property
    def deepsky_forecast_tomorrow_dayname(self):
        """Return Forecast Todays Dayname."""
        if len(self._deepsky_forecast) > 1:
            nightly_conditions = self._deepsky_forecast[1]
            return nightly_conditions.dayname
        return None

    @property
    def deepsky_forecast_tomorrow_plain(self):
        """Return Forecast Tomorrow."""
        out = ""
        for nightly_condition in self._deepsky_forecast[1].nightly_conditions:
            out += CONDITION[4 - math.floor(nightly_condition / 20)].capitalize()
        return out

    @property
    def deepsky_forecast_tomorrow_desc(self):
        """Return Forecast Tomorrow Description."""
        if len(self._deepsky_forecast) > 1:
            nightly_conditions = self._deepsky_forecast[1]
            return nightly_conditions.weather.replace("_", " ").capitalize()
        return None

    @property
    def deepsky_forecast(self):
        """Return Deepsky Forecast."""
        return self._deepsky_forecast

    @property
    def uptonight(self) -> int:
        """Return the number of best DSOs for tonight."""
        if self._uptonight is not None:
            return len(self._uptonight)
        return None

    @property
    def uptonight_list(self) -> []:
        if self._uptonight is not None:
            return self._uptonight
    
class ForecastData(BaseData):
    """A representation of 3-Hour Based Forecast AstroWeather Data."""

    def __init__(self, data):
        super().__init__(data)
        self._hour = data["hour"]

    @property
    def hour(self) -> int:
        """Return Forecast Hour of the day."""
        return self._hour

    @property
    def deep_sky_view(self) -> bool:
        """Return True if Deep Sky should be possible."""
        if self.condition_percentage <= DEEP_SKY_THRESHOLD:
            return True
        return False


class NightlyConditionsData:
    """A representation of nights Sky Quality Data."""

    def __init__(self, data):
        self._init = data["init"]
        self._dayname = data["dayname"]
        self._hour = data["hour"]
        self._nightly_conditions = data["nightly_conditions"]
        self._weather = data["weather"]

    @property
    def init(self) -> datetime:
        """Return Forecast Anchor."""
        return self._init

    @property
    def dayname(self) -> str:
        """Return Forecast Name of the Day."""
        return self._dayname

    @property
    def hour(self) -> int:
        """Return Forecast Hour."""
        return self._hour

    @property
    def nightly_conditions(self) -> int:
        """Return Forecast Hour."""
        return self._nightly_conditions

    @property
    def weather(self) -> str:
        """Return Current Weather."""
        return self._weather.replace("_", " ").capitalize()

class DSOUpTonight:
    """A representation of uptonight DSO."""
    
    def __init__(self, data):
        self._target_name = data["target_name"]
        self._type = data["type"]
        self._constellation = data["constellation"]
        self._size = data["size"]
        self._foto = data["foto"]
    
    @property
    def target_name(self) -> str:
        """Return Forecast Name of the Day."""
        return self._target_name

    @property
    def type(self) -> str:
        """Return Forecast Name of the Day."""
        return self._type

    @property
    def constellation(self) -> str:
        """Return Forecast Name of the Day."""
        return self._constellation

    @property
    def size(self) -> str:
        """Return Forecast Name of the Day."""
        return self._size

    @property
    def foto(self) -> str:
        """Return Forecast Name of the Day."""
        return self._foto



""" Contains Helper functions for AstroWeather."""
from datetime import datetime, timedelta
import logging
import ephem
from ephem import degree
from math import degrees as deg
import pytz

# The introduction of the module [timezonefinder](https://github.com/jannikmi/timezonefinder)
# with it's nested dependency to [py-h3](https://github.com/uber/h3-py) failed while compiling
# the `c`-module h3 on some home assistant deployment variants (e.g. Home Assistant
# Operating System on RPi).
# from timezonefinder import TimezoneFinder
from pyastroweatherio.const import (
    DEFAULT_ELEVATION,
    DEFAULT_TIMEZONE,
    HOME_LATITUDE,
    HOME_LONGITUDE,
    CIVIL_TWILIGHT,
    CIVIL_DUSK_DAWN,
    NAUTICAL_TWILIGHT,
    NAUTICAL_DUSK_DAWN,
    ASTRONOMICAL_TWILIGHT,
    ASTRONOMICAL_DUSK_DAWN,
)

_LOGGER = logging.getLogger(__name__)


class ConversionFunctions:
    """Convert between different Weather Units."""

    async def epoch_to_datetime(self, value) -> str:
        """Converts EPOC time to Date Time String."""
        return datetime.datetime.fromtimestamp(int(value)).strftime("%Y-%m-%d %H:%M:%S")

    async def anchor_timestamp(self, value) -> datetime:
        """Converts the datetime string from 7Timer to DateTime."""
        return datetime.strptime(value, "%Y%m%d%H")


class AstronomicalRoutines:
    """Calculate different astronomical objects"""

    def __init__(
        self,
        latitude=HOME_LATITUDE,
        longitude=HOME_LONGITUDE,
        elevation=DEFAULT_ELEVATION,
        timezone_info=DEFAULT_TIMEZONE,
        forecast_time=None,
    ):
        self._latitude = latitude
        self._longitude = longitude
        self._elevation = elevation
        self._timezone_info = timezone_info
        # tz_find = TimezoneFinder()
        # self._timezone_info = tz_find.timezone_at(lng=longitude, lat=latitude)
        self._test_mode = False
        _LOGGER.debug("Timezone: %s", self._timezone_info)
        if forecast_time is None:
            self._forecast_time = datetime.utcnow().replace(tzinfo=pytz.utc)
        else:
            self._forecast_time = forecast_time.replace(tzinfo=pytz.utc)
            self._test_mode = True
        _LOGGER.debug("Forecast Time: %s", self._forecast_time)

        self._sun_observer = None
        self._sun_observer_nautical = None
        self._sun_observer_astro = None
        self._moon_observer = None
        self._sun_next_rising_civil = None
        self._sun_next_setting_civil = None
        self._sun_next_rising_nautical = None
        self._sun_next_setting_nautical = None
        self._sun_next_rising_astro = None
        self._sun_next_setting_astro = None
        self._sun_altitude = None
        self._sun_azimuth = None
        self._moon_next_rising = None
        self._moon_next_setting = None
        self._moon_next_new_moon = None
        self._moon_next_full_moon = None
        self._moon_altitude = None
        self._moon_azimuth = None
        self._sun = None
        self._moon = None

        # Internal only
        self._sun_previous_rising_astro = None
        self._sun_previous_setting_astro = None
        self._moon_previous_rising = None
        self._moon_previous_setting = None
        # self._moon_day_after_next_rising = None
        # self._moon_day_after_next_setting = None

    def utc_to_local(self, utc_dt):
        """Localizes the datetime"""
        local_tz = pytz.timezone(self._timezone_info)
        local_dt = utc_dt.replace(tzinfo=pytz.utc).astimezone(local_tz)
        return local_tz.normalize(local_dt)

    def utc_to_local_diff(self):
        """returns the UTC Offset"""
        local = self.utc_to_local(datetime.now())
        return local.utcoffset().seconds // 3600

    #
    # Observers
    #
    def get_sun_observer(self, below_horizon=ASTRONOMICAL_DUSK_DAWN) -> ephem.Observer:
        """Retrieves the ephem sun observer for the current location"""
        observer = ephem.Observer()
        observer.lon = self._longitude * degree
        observer.lat = self._latitude * degree
        observer.elevation = self._elevation
        observer.horizon = below_horizon * degree
        observer.pressure = 0
        observer.epoch = datetime.now().strftime("%Y/%m/%d")
        return observer

    def get_moon_observer(self) -> ephem.Observer:
        """Retrieves the ephem mon observer for the current location"""
        observer = ephem.Observer()
        observer.lon = self._longitude * degree
        observer.lat = self._latitude * degree
        observer.elevation = self._elevation
        # Naval Observatory Risings and Settings
        # Set horizon to minus 34 arcminutes
        # https://aa.usno.navy.mil/data/RS_OneDay
        observer.horizon = "-0:34"
        observer.pressure = 0
        observer.epoch = datetime.now().strftime("%Y/%m/%d")
        return observer

    #
    # Sun & Moon
    #
    def calculate_sun(self):
        """Calculates sun risings and settings"""
        if self._sun_observer is None:
            self._sun_observer = self.get_sun_observer(CIVIL_DUSK_DAWN)
        if self._sun_observer_nautical is None:
            self._sun_observer_nautical = self.get_sun_observer(NAUTICAL_DUSK_DAWN)
        if self._sun_observer_astro is None:
            self._sun_observer_astro = self.get_sun_observer(ASTRONOMICAL_DUSK_DAWN)
        if self._sun is None:
            self._sun = ephem.Sun()

        # Rise and Setting (Civil)
        try:
            self._sun_next_rising_civil = (
                self._sun_observer.next_rising(ephem.Sun(), use_center=True).datetime().replace(tzinfo=pytz.utc)
            )
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next rising
            start = self._sun_observer.date.datetime()
            end = self._sun_observer.date.datetime() + timedelta(days=365)
            timestamp = start
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer.date = timestamp
                try:
                    self._sun_next_rising_astro = self._sun_observer.next_rising(
                        ephem.Sun(), use_center=True
                    ).datetime()
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

        try:
            self._sun_next_setting_civil = self._sun_observer.next_setting(ephem.Sun(), use_center=True).datetime().replace(tzinfo=pytz.utc)
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next setting
            start = self._sun_observer.date.datetime()
            end = self._sun_observer.date.datetime() + timedelta(days=365)
            timestamp = start
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer.date = timestamp
                try:
                    self._sun_next_setting_astro = self._sun_observer.next_setting(
                        ephem.Sun(), use_center=True
                    ).datetime()
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

        # Rise and Setting (Nautical)
        self._sun_observer_nautical.date = self._forecast_time
        self._sun.compute(self._sun_observer_nautical)

        try:
            self._sun_next_rising_nautical = self._sun_observer_nautical.next_rising(
                ephem.Sun(), use_center=True
            ).datetime().replace(tzinfo=pytz.utc)
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical rising
            start = self._sun_observer_nautical.date.datetime()
            end = self._sun_observer_nautical.date.datetime() + timedelta(days=365)
            timestamp = start
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer_nautical.date = timestamp
                try:
                    self._sun_next_rising_nautical = self._sun_observer_nautical.next_rising(
                        ephem.Sun(), use_center=True
                    ).datetime().replace(tzinfo=pytz.utc)
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

        try:
            self._sun_next_setting_nautical = self._sun_observer_nautical.next_setting(
                ephem.Sun(), use_center=True
            ).datetime().replace(tzinfo=pytz.utc)
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical setting
            start = self._sun_observer_nautical.date.datetime()
            end = self._sun_observer_nautical.date.datetime() + timedelta(days=365)
            timestamp = start
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer_nautical.date = timestamp
                try:
                    self._sun_next_setting_nautical = self._sun_observer_nautical.next_setting(
                        ephem.Sun(), use_center=True
                    ).datetime().replace(tzinfo=pytz.utc)
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

        # Rise and Setting (Astronomical)
        self._sun_observer_astro.date = self._forecast_time
        self._sun.compute(self._sun_observer_astro)

        try:
            self._sun_next_rising_astro = self._sun_observer_astro.next_rising(ephem.Sun(), use_center=True).datetime().replace(tzinfo=pytz.utc)
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical rising
            start = self._sun_observer_astro.date.datetime()
            end = self._sun_observer_astro.date.datetime() + timedelta(days=365)
            timestamp = start
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer_astro.date = timestamp
                try:
                    self._sun_next_rising_astro = self._sun_observer_astro.next_rising(
                        ephem.Sun(), use_center=True
                    ).datetime().replace(tzinfo=pytz.utc)
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

        try:
            self._sun_previous_rising_astro = self._sun_observer_astro.previous_rising(
                ephem.Sun(), use_center=True
            ).datetime().replace(tzinfo=pytz.utc)
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the previous astronomical rising
            start = self._sun_observer_astro.date.datetime()
            end = self._sun_observer_astro.date.datetime() - timedelta(days=365)
            timestamp = start
            while timestamp > end:
                timestamp -= timedelta(minutes=1440)
                self._sun_observer_astro.date = timestamp
                try:
                    self._sun_previous_rising_astro = self._sun_observer_astro.previous_rising(
                        ephem.Sun(), use_center=True
                    ).datetime().replace(tzinfo=pytz.utc)
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

        try:
            self._sun_next_setting_astro = self._sun_observer_astro.next_setting(
                ephem.Sun(), use_center=True
            ).datetime().replace(tzinfo=pytz.utc)
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the next astronomical setting
            start = self._sun_observer_astro.date.datetime()
            end = self._sun_observer_astro.date.datetime() + timedelta(days=365)
            timestamp = start
            while timestamp < end:
                timestamp += timedelta(minutes=1440)
                self._sun_observer_astro.date = timestamp
                try:
                    self._sun_next_setting_astro = self._sun_observer_astro.next_setting(
                        ephem.Sun(), use_center=True
                    ).datetime().replace(tzinfo=pytz.utc)
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

        try:
            self._sun_previous_setting_astro = self._sun_observer_astro.previous_setting(
                ephem.Sun(), use_center=True
            ).datetime().replace(tzinfo=pytz.utc)
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            # Search for the previous astronomical setting
            start = self._sun_observer_astro.date.datetime()
            end = self._sun_observer_astro.date.datetime() - timedelta(days=365)
            timestamp = start
            while timestamp > end:
                timestamp -= timedelta(minutes=1440)
                self._sun_observer_astro.date = timestamp
                try:
                    self._sun_previous_setting_astro = self._sun_observer_astro.previous_setting(
                        ephem.Sun(), use_center=True
                    ).datetime().replace(tzinfo=pytz.utc)
                except (ephem.AlwaysUpError, ephem.NeverUpError):
                    continue
                break

    def calculate_sun_altaz(self):
        """Calculates sun altitude and azimuth"""
        if self._sun_observer is None:
            self._sun_observer = self.get_sun_observer(CIVIL_DUSK_DAWN)
        if self._sun is None:
            self._sun = ephem.Sun()

        self._sun_observer.date = self._forecast_time
        self._sun.compute(self._sun_observer)

        # Sun Altitude
        self._sun_altitude = deg(float(self._sun.alt))

        # Sun Azimuth
        self._sun_azimuth = deg(float(self._sun.az))

    def calculate_moon(self):
        """Calculates moon rising and setting"""
        if self._moon_observer is None:
            self._moon_observer = self.get_moon_observer()
        if self._moon is None:
            self._moon = ephem.Moon()

        # Rise and Setting
        self._moon_observer.date = self._forecast_time
        self._moon.compute(self._moon_observer)

        try:
            self._moon_next_rising = self._moon_observer.next_rising(ephem.Moon()).datetime().replace(tzinfo=pytz.utc)
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            pass

        try:
            self._moon_next_setting = self._moon_observer.next_setting(ephem.Moon()).datetime().replace(tzinfo=pytz.utc)
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            pass

        try:
            self._moon_previous_rising = self._moon_observer.previous_rising(ephem.Moon()).datetime().replace(tzinfo=pytz.utc)
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            pass

        try:
            self._moon_previous_setting = self._moon_observer.previous_setting(ephem.Moon()).datetime().replace(tzinfo=pytz.utc)
        except (ephem.AlwaysUpError, ephem.NeverUpError):
            pass
        
        # self._moon_observer.date = self._forecast_time + timedelta(days=1)
        # self._moon.compute(self._moon_observer)

        # try:
        #     self._moon_day_after_next_rising = self._moon_observer.next_rising(ephem.Moon()).datetime().replace(tzinfo=pytz.utc)
        # except (ephem.AlwaysUpError, ephem.NeverUpError):
        #     pass

        # try:
        #     self._moon_day_after_next_setting = self._moon_observer.next_setting(ephem.Moon()).datetime().replace(tzinfo=pytz.utc)
        # except (ephem.AlwaysUpError, ephem.NeverUpError):
        #     pass

        # Next new Moon
        self._moon_next_new_moon = ephem.next_new_moon(self._forecast_time).datetime().replace(tzinfo=pytz.utc)

        # Next full Moon
        self._moon_next_full_moon = ephem.next_full_moon(self._forecast_time).datetime().replace(tzinfo=pytz.utc)

    def calculate_moon_altaz(self):
        """Calculates moon altitude and azimuth"""
        if self._moon_observer is None:
            self._moon_observer = self.get_moon_observer()
        if self._moon is None:
            self._moon = ephem.Moon()

        self._moon_observer.date = self._forecast_time
        self._moon.compute(self._moon_observer)

        # Moon Altitude
        self._moon_altitude = deg(float(self._moon.alt))

        # Moon Azimuth
        self._moon_azimuth = deg(float(self._moon.az))

    def sun_previous_rising_astro(self) -> datetime:
        """Returns sun previous astronomical rising"""
        if self._sun_previous_rising_astro is None or self._forecast_time > self._sun_previous_rising_astro:
            _LOGGER.debug("Astronomical calculations updating sun_previous_rising_astro")
            self.calculate_sun()

        if self._sun_previous_rising_astro is not None:
            return self._sun_previous_rising_astro

    def sun_previous_setting_astro(self) -> datetime:
        """Returns sun previous astronomical setting"""
        if self._sun_previous_setting_astro is None or self._forecast_time > self._sun_previous_setting_astro:
            _LOGGER.debug("Astronomical calculations updating sun_previous_setting_astro")
            self.calculate_sun()

        if self._sun_previous_setting_astro is not None:
            return self._sun_previous_setting_astro

    def astronomical_darkness(self) -> bool:
        """Returns true during astronomical night"""
        
        if self._sun_next_setting_astro > self._sun_next_rising_astro:
            return True
        return False
    
    def moon_down(self) -> bool:
        """Returns true while moon is set"""
        
        if self._moon_next_setting > self._moon_next_rising:
            return True
        return False

    #
    # Public methods
    #
    async def need_update(self):
        """Update Sun and Moon"""
        self.calculate_sun_altaz()
        self.calculate_moon_altaz()

        if (
            self._sun_next_setting_civil is None
            or self._sun_next_setting_nautical is None
            or self._sun_next_setting_astro is None
            or self._sun_next_rising_astro is None
            or self._sun_next_rising_nautical is None
            or self._sun_next_rising_civil is None
            or self._moon_next_rising is None
            or self._moon_next_setting is None
            or self._forecast_time > self._sun_next_setting_civil
            or self._forecast_time > self._sun_next_setting_nautical
            or self._forecast_time > self._sun_next_setting_astro
            or self._forecast_time > self._sun_next_rising_astro
            or self._forecast_time > self._sun_next_rising_nautical
            or self._forecast_time > self._sun_next_rising_civil
            or self._forecast_time > self._moon_next_rising
            or self._forecast_time > self._moon_next_setting
        ):
            _LOGGER.debug("Astronomical calculations updating")
            self.calculate_sun()
            self.calculate_moon()

    # Return Sun information
    async def sun_next_rising(self) -> datetime:
        """Returns sun next rising"""
        if (
            self._sun_next_rising_astro is None
            or self._sun_next_rising_nautical is None
            or self._sun_next_rising_civil is None
            or self._forecast_time > self._sun_next_rising_astro
            or self._forecast_time > self._sun_next_rising_nautical
            or self._forecast_time > self._sun_next_rising_civil
        ):
            _LOGGER.debug("Astronomical calculations updating sun_next_rising")
            self.calculate_sun()

        if self._sun_next_rising_astro is not None:
            return self._sun_next_rising_astro
        if self._sun_next_rising_nautical is not None:
            return self._sun_next_rising_nautical
        if self._sun_next_rising_civil is not None:
            return self._sun_next_rising_civil

    async def sun_next_rising_civil(self) -> datetime:
        """Returns sun next rising"""
        if self._sun_next_setting_civil is None or self._forecast_time > self._sun_next_setting_civil:
            _LOGGER.debug("Astronomical calculations updating sun_next_rising_civil")
            self.calculate_sun()

        if self._sun_next_rising_civil is not None:
            return self._sun_next_rising_civil

    async def sun_next_rising_nautical(self) -> datetime:
        """Returns sun next nautical rising"""
        if self._sun_next_rising_nautical is None or self._forecast_time > self._sun_next_rising_nautical:
            _LOGGER.debug("Astronomical calculations updating sun_next_rising_nautical")
            self.calculate_sun()

        if self._sun_next_rising_nautical is not None:
            return self._sun_next_rising_nautical

    async def sun_next_rising_astro(self) -> datetime:
        """Returns sun next astronomical rising"""
        if self._sun_next_rising_astro is None or self._forecast_time > self._sun_next_rising_astro:
            _LOGGER.debug("Astronomical calculations updating sun_next_rising_astro")
            self.calculate_sun()

        if self._sun_next_rising_astro is not None:
            return self._sun_next_rising_astro

    async def sun_next_setting(self) -> datetime:
        """Returns sun next setting"""
        if (
            self._sun_next_setting_civil is None
            or self._sun_next_setting_nautical is None
            or self._sun_next_setting_astro is None
            or self._forecast_time > self._sun_next_setting_civil
            or self._forecast_time > self._sun_next_setting_nautical
            or self._forecast_time > self._sun_next_setting_astro
        ):
            _LOGGER.debug("Astronomical calculations updating sun_next_setting")
            self.calculate_sun()

        if self._sun_next_setting_astro is not None:
            return self._sun_next_setting_astro
        if self._sun_next_setting_nautical is not None:
            return self._sun_next_setting_nautical
        if self._sun_next_setting_civil is not None:
            return self._sun_next_setting_civil

    async def sun_next_setting_civil(self) -> datetime:
        """Returns sun next setting"""
        if self._sun_next_setting_civil is None or self._forecast_time > self._sun_next_setting_civil:
            _LOGGER.debug("Astronomical calculations updating sun_next_setting_civil")
            self.calculate_sun()

        if self._sun_next_setting_civil is not None:
            return self._sun_next_setting_civil

    async def sun_next_setting_nautical(self) -> datetime:
        """Returns sun next nautical setting"""
        if self._sun_next_setting_nautical is None or self._forecast_time > self._sun_next_setting_nautical:
            _LOGGER.debug("Astronomical calculations updating sun_next_setting_nautical")
            self.calculate_sun()

        if self._sun_next_setting_nautical is not None:
            return self._sun_next_setting_nautical

    async def sun_next_setting_astro(self) -> datetime:
        """Returns sun next astronomical setting"""
        if self._sun_next_setting_astro is None or self._forecast_time > self._sun_next_setting_astro:
            _LOGGER.debug("Astronomical calculations updating sun_next_setting_astro")
            self.calculate_sun()

        if self._sun_next_setting_astro is not None:
            return self._sun_next_setting_astro

    async def sun_altitude(self) -> float:
        """Returns the sun altitude"""
        self.calculate_sun_altaz()

        if self._sun_altitude is not None:
            return self._sun_altitude

    async def sun_azimuth(self) -> float:
        """Returns the sun azimuth"""
        self.calculate_sun_altaz()

        if self._sun_azimuth is not None:
            return self._sun_azimuth

    # Return Moon information
    async def moon_next_rising(self) -> datetime:
        if self._moon_next_rising is None or self._forecast_time > self._moon_next_rising:
            _LOGGER.debug("Astronomical calculations updating moon_next_rising")
            self.calculate_moon()

        """Returns moon next rising"""
        if self._moon_next_rising is not None:
            return self._moon_next_rising

    async def moon_next_setting(self) -> datetime:
        if self._moon_next_setting is None or self._forecast_time > self._moon_next_setting:
            _LOGGER.debug("Astronomical calculations updating moon_next_setting")
            self.calculate_moon()

        """Returns moon next setting"""
        if self._moon_next_setting is not None:
            return self._moon_next_setting

    async def moon_phase(self) -> float:
        """Returns the moon phase"""
        self.calculate_moon()

        if self._moon is not None:
            return self._moon.phase

    async def moon_next_new_moon(self) -> float:
        """Returns the next new moon"""
        if self._moon_next_new_moon is None or self._forecast_time > self._moon_next_new_moon:
            _LOGGER.debug("Astronomical calculations updating moon_next_new_moon")
            self.calculate_moon()

        if self._moon_next_new_moon is not None:
            return self._moon_next_new_moon

    async def moon_next_full_moon(self) -> float:
        """Returns the next full moon"""
        if self._moon_next_full_moon is None or self._forecast_time > self._moon_next_full_moon:
            _LOGGER.debug("Astronomical calculations updating moon_next_full_moon")
            self.calculate_moon()

        if self._moon_next_full_moon is not None:
            return self._moon_next_full_moon

    async def moon_altitude(self) -> float:
        """Returns the moon altitude"""
        self.calculate_moon_altaz()

        if self._moon_altitude is not None:
            return self._moon_altitude

    async def moon_azimuth(self) -> float:
        """Returns the moon azimuth"""
        self.calculate_moon_altaz()

        if self._moon_azimuth is not None:
            return self._moon_azimuth

    # Astronomical Night and Darkness information
    async def night_duration_astronomical(self) -> float:
        """Returns the remaining timespan of astronomical darkness"""
        start_timestamp = None

        # Are we already in darkness?
        if self.astronomical_darkness():
            start_timestamp = self._sun_previous_setting_astro
        else:
            start_timestamp = self._sun_next_setting_astro

        astroduration = self._sun_next_rising_astro - start_timestamp

        return astroduration.total_seconds()
    
    async def deep_sky_darkness_moon_rises(self) -> bool:
        """Returns true if moon rises during astronomical night"""
        start_timestamp = None

        # Are we already in darkness?
        if self.astronomical_darkness():
            start_timestamp = self._sun_previous_setting_astro
        else:
            start_timestamp = self._sun_next_setting_astro

        if self._moon_next_rising > start_timestamp and self._moon_next_rising < self._sun_next_rising_astro:
            _LOGGER.debug("Moon rises during astronomical night")
            return True
        return False

    async def deep_sky_darkness_moon_sets(self) -> bool:
        """Returns true if moon sets during astronomical night"""
        start_timestamp = None

        # Are we already in darkness?
        if self.astronomical_darkness():
            start_timestamp = self._sun_previous_setting_astro
        else:
            start_timestamp = self._sun_next_setting_astro

        # Did Moon already set in darkness?
        if self.moon_down() and self.astronomical_darkness():
            start_timestamp_moon = self._moon_previous_setting
        else:
            start_timestamp_moon = self._moon_next_setting
        
        if start_timestamp_moon > start_timestamp and start_timestamp_moon < self._sun_next_rising_astro:
            _LOGGER.debug("Moon sets during astronomical night")
            return True
        return False

    async def deep_sky_darkness_moon_always_up(self) -> bool:
        """Returns true if moon is up during astronomical night"""
        start_timestamp = None

        # Are we already in darkness?
        if self.astronomical_darkness():
            start_timestamp = self._sun_previous_setting_astro
        else:
            start_timestamp = self._sun_next_setting_astro

        if self._moon_next_rising < start_timestamp and self._moon_next_setting > self._sun_next_rising_astro:
            _LOGGER.debug("Moon is up during astronomical night")
            return True
        return False

    async def deep_sky_darkness_moon_always_down(self) -> bool:
        """Returns true if moon is down during astronomical night"""
        start_timestamp = None

        # Are we already in darkness?
        if self.astronomical_darkness():
            start_timestamp = self._sun_previous_setting_astro
        else:
            start_timestamp = self._sun_next_setting_astro

        if self._moon_next_setting < start_timestamp and self._moon_next_rising > self._sun_next_rising_astro:
            _LOGGER.debug("Moon is up during astronomical night")
            return True
        return False
        
    async def deep_sky_darkness(self) -> float:
        """Returns the remaining timespan of deep sky darkness"""
        dsd = timedelta(0)

        if self.astronomical_darkness():
            if await self.deep_sky_darkness_moon_rises():
                dsd = self._moon_next_rising - self._forecast_time
                _LOGGER.debug(f"DSD - Sun down, Moon rises {dsd}")

            if await self.deep_sky_darkness_moon_sets():
                if self.moon_down():
                    dsd = self._sun_next_rising_astro - self._forecast_time
                    _LOGGER.debug(f"DSD - Sun down, Moon is down {dsd}")
                else:
                    dsd = self._sun_next_rising_astro - self._moon_next_setting
                    _LOGGER.debug(f"DSD - Sun down, Moon sets {dsd}")

            if await self.deep_sky_darkness_moon_always_down():
                dsd = self._sun_next_rising_astro - self._forecast_time
                _LOGGER.debug(f"DSD - Sun down, Moon down {dsd}")
            
        if not self.astronomical_darkness():
            if await self.deep_sky_darkness_moon_rises():
                dsd = self._moon_next_rising - self._sun_next_setting_astro
                _LOGGER.debug(f"DSD - Sun up, Moon rises {dsd}")

            if await self.deep_sky_darkness_moon_sets():
                dsd = self._sun_next_rising_astro - self._moon_next_setting
                _LOGGER.debug(f"DSD - Sun up, Moon sets {dsd}")

            if await self.deep_sky_darkness_moon_always_down():
                dsd = self._sun_next_rising_astro - self._sun_next_setting_astro
                _LOGGER.debug(f"DSD - Sun up, Moon down {dsd}")
            
        if await self.deep_sky_darkness_moon_always_up():
            dsd = timedelta(0)

        return dsd.total_seconds()



"""Define a client to interact with 7Timer."""
import asyncio
import json
import logging
import os.path
from datetime import datetime, timedelta
from typing import Optional
from decimal import Decimal
from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError

from pyastroweatherio.const import (
    BASE_URL_SEVENTIMER,
    BASE_URL_MET,
    DEFAULT_TIMEOUT,
    DEFAULT_CACHE_TIMEOUT,
    DEFAULT_ELEVATION,
    DEFAULT_TIMEZONE,
    DEFAULT_CONDITION_CLOUDCOVER_WEIGHT,
    DEFAULT_CONDITION_SEEING_WEIGHT,
    DEFAULT_CONDITION_TRANSPARENCY_WEIGHT,
    HOME_LATITUDE,
    HOME_LONGITUDE,
    STIMER_OUTPUT,
    # FORECAST_TYPE_DAILY,
    FORECAST_TYPE_HOURLY,
    MAGNUS_COEFFICIENT_A,
    MAGNUS_COEFFICIENT_B,
)
from pyastroweatherio.dataclasses import (
    ForecastData,
    LocationData,
    NightlyConditionsData,
    DSOUpTonight,
)
from pyastroweatherio.errors import RequestError
from pyastroweatherio.helper_functions import ConversionFunctions, AstronomicalRoutines

_LOGGER = logging.getLogger(__name__)


class AstroWeather:
    """AstroWeather Communication Client."""

    def __init__(
        self,
        session: Optional[ClientSession] = None,
        latitude=HOME_LATITUDE,
        longitude=HOME_LONGITUDE,
        elevation=DEFAULT_ELEVATION,
        timezone_info=DEFAULT_TIMEZONE,
        cloudcover_weight=DEFAULT_CONDITION_CLOUDCOVER_WEIGHT,
        seeing_weight=DEFAULT_CONDITION_SEEING_WEIGHT,
        transparency_weight=DEFAULT_CONDITION_TRANSPARENCY_WEIGHT,
        uptonight_path="/conf/www",
        test_datetime=None,
    ):
        self._session: ClientSession = session
        self._latitude = latitude
        self._longitude = longitude
        self._elevation = elevation
        self._timezone_info = timezone_info
        self._weather_data_seventimer = []
        self._weather_data_seventimer_init = ""
        self._weather_data_metno = []
        self._weather_data_metno_init = ""
        self._weather_data_uptonight = {}
        self._weather_data_seventimer_timestamp = datetime.now() - timedelta(seconds=(DEFAULT_CACHE_TIMEOUT + 1))
        self._weather_data_metno_timestamp = datetime.now() - timedelta(seconds=(DEFAULT_CACHE_TIMEOUT + 1))
        self._data_uptonight_timestamp = datetime.now() - timedelta(seconds=(DEFAULT_CACHE_TIMEOUT + 1))
        self._cloudcover_weight = cloudcover_weight
        self._seeing_weight = seeing_weight
        self._transparency_weight = transparency_weight
        self._uptonight_path = uptonight_path
        self._test_datetime = test_datetime

        self._forecast_data = None

        self.req = session

        # Astro Routines
        self._astro_routines = AstronomicalRoutines(
            self._latitude, self._longitude, self._elevation, self._timezone_info, self._test_datetime
        )

        # Testing
        self.test_mode = False
        self.dump_json = False
        if self._test_datetime is not None:
            self.test_mode = True
            self.dump_json = True

    # Public functions
    async def get_location_data(
        self,
    ) -> None:
        """Returns station Weather Forecast."""
        return await self._get_location_data()

    # async def get_forecast(self, forecast_type=FORECAST_TYPE_DAILY, hours_to_show=24) -> None:
    #     """Returns station Weather Forecast."""
    #     _LOGGER.debug("get_forecast called")
    #     return await self._get_forecast_data(forecast_type, hours_to_show)

    # async def get_daily_forecast(self) -> None:
    #     """Returns daily Weather Forecast."""
    #     return await self._get_forecast_data(FORECAST_TYPE_DAILY, 72)

    async def get_hourly_forecast(self) -> None:
        """Returns hourly Weather Forecast."""
        return await self._get_forecast_data(FORECAST_TYPE_HOURLY, 72)

    async def get_deepsky_forecast(self) -> None:
        """Returns Deep Sky Forecast."""
        return await self._get_deepsky_forecast()

    # Private functions
    async def _get_location_data(self) -> None:
        """Return Forecast data"""

        cnv = ConversionFunctions()
        items = []

        await self.retrieve_data_seventimer()
        await self.retrieve_data_metno()
        now = datetime.utcnow()

        # Anchor timestamp
        init_ts = await cnv.anchor_timestamp(self._weather_data_seventimer_init)

        await self._astro_routines.need_update()

        # Met.no
        metno_index = -1
        forecast_skipped = 0
        for row in self._weather_data_seventimer:
            # 7Timer: Skip over past forecasts
            forecast_time = init_ts + timedelta(hours=row["timepoint"])
            if now > forecast_time:
                forecast_skipped += 1
                continue

            _LOGGER.debug("7Timer forecast time: %s", str(forecast_time))

            # Met.no: Search for 7Timer forecast time
            for datapoint in self._weather_data_metno:
                metno_index += 1
                if forecast_time == datetime.strptime(datapoint.get("time"), "%Y-%m-%dT%H:%M:%SZ"):
                    break
            _LOGGER.debug("Met.no start index: %s", str(metno_index))

            item = {
                "init": init_ts,
                "timepoint": row["timepoint"],
                "timestamp": forecast_time,
                "forecast_length": (len(self._weather_data_seventimer) - forecast_skipped) * 3,
                "latitude": self._latitude,
                "longitude": self._longitude,
                "elevation": self._elevation,
                "seeing": row["seeing"],
                "transparency": row["transparency"],
                "lifted_index": row["lifted_index"],
                "sun_next_rising": await self._astro_routines.sun_next_rising_civil(),
                "sun_next_rising_nautical": await self._astro_routines.sun_next_rising_nautical(),
                "sun_next_rising_astro": await self._astro_routines.sun_next_rising_astro(),
                "sun_next_setting": await self._astro_routines.sun_next_setting_civil(),
                "sun_next_setting_nautical": await self._astro_routines.sun_next_setting_nautical(),
                "sun_next_setting_astro": await self._astro_routines.sun_next_setting_astro(),
                "sun_altitude": await self._astro_routines.sun_altitude(),
                "sun_azimuth": await self._astro_routines.sun_azimuth(),
                "moon_next_rising": await self._astro_routines.moon_next_rising(),
                "moon_next_setting": await self._astro_routines.moon_next_setting(),
                "moon_phase": await self._astro_routines.moon_phase(),
                "moon_next_new_moon": await self._astro_routines.moon_next_new_moon(),
                "moon_next_full_moon": await self._astro_routines.moon_next_full_moon(),
                "moon_altitude": await self._astro_routines.moon_altitude(),
                "moon_azimuth": await self._astro_routines.moon_azimuth(),
                "night_duration_astronomical": await self._astro_routines.night_duration_astronomical(),
                "deep_sky_darkness_moon_rises": await self._astro_routines.deep_sky_darkness_moon_rises(),
                "deep_sky_darkness_moon_sets": await self._astro_routines.deep_sky_darkness_moon_sets(),
                "deep_sky_darkness_moon_always_up": await self._astro_routines.deep_sky_darkness_moon_always_up(),
                "deep_sky_darkness_moon_always_down": await self._astro_routines.deep_sky_darkness_moon_always_down(),
                "deep_sky_darkness": await self._astro_routines.deep_sky_darkness(),
                "deepsky_forecast": await self._get_deepsky_forecast(),
                "uptonight": await self._get_deepsky_objects(),
            }
            # Met.no
            if (
                datetime.strptime(
                    self._weather_data_metno[metno_index].get("time"),
                    "%Y-%m-%dT%H:%M:%SZ",
                )
                == forecast_time
            ):
                # _LOGGER.debug("Met.no Cloud Area Fraction timestamp match: %s", str(forecast_time))
                details = self._weather_data_metno[metno_index].get("data", {}).get("instant", {}).get("details", {})
                item["cloudcover"] = int(details.get("cloud_area_fraction", -1) / 12.5 + 1)

                item["cloud_area_fraction"] = details.get("cloud_area_fraction", -1)
                item["cloud_area_fraction_high"] = details.get("cloud_area_fraction_high", -1)
                item["cloud_area_fraction_low"] = details.get("cloud_area_fraction_low", -1)
                item["cloud_area_fraction_medium"] = details.get("cloud_area_fraction_medium", -1)
                item["fog_area_fraction"] = details.get("fog_area_fraction", -1)

                item["condition_percentage"] = await self.calc_condition_percentage(
                    item["cloud_area_fraction"] / 12.5 + 1,
                    row["seeing"],
                    row["transparency"],
                )

                item["rh2m"] = details.get("relative_humidity", -1)
                item["wind_speed"] = details.get("wind_speed", -1)
                item["wind_from_direction"] = details.get("wind_from_direction", -1)
                item["temp2m"] = details.get("air_temperature", -1)
                item["dewpoint2m"] = details.get("dew_point_temperature", -1)
                item["weather"] = (
                    self._weather_data_metno[metno_index]
                    .get("data", {})
                    .get("next_1_hours", {})
                    .get("summary", {})
                    .get("symbol_code", "")
                )
                item["weather6"] = (
                    self._weather_data_metno[metno_index]
                    .get("data", {})
                    .get("next_6_hours", {})
                    .get("summary", {})
                    .get("symbol_code", "")
                )
                item["precipitation_amount"] = (
                    self._weather_data_metno[metno_index]
                    .get("data", {})
                    .get("next_1_hours", {})
                    .get("details", {})
                    .get("precipitation_amount", "")
                )

            else:
                break

            items.append(LocationData(item))
            break

        return items

    async def _get_forecast_data(self, forecast_type, hours_to_show) -> None:
        """Return Forecast data for the Station."""

        cnv = ConversionFunctions()
        items = []

        await self.retrieve_data_seventimer()
        await self.retrieve_data_metno()
        now = datetime.utcnow()

        # Create items
        cnt = 0

        # Anchor timestamp
        init_ts = await cnv.anchor_timestamp(self._weather_data_seventimer_init)

        utc_to_local_diff = self._astro_routines.utc_to_local_diff()
        _LOGGER.debug("UTC to local diff: %s", str(utc_to_local_diff))
        _LOGGER.debug("Forecast length: %s", str(len(self._weather_data_seventimer)))

        # Met.no
        metno_index = -1
        for row in self._weather_data_seventimer:
            # 7Timer: Skip over past forecasts
            forecast_time = init_ts + timedelta(hours=row["timepoint"])
            if now > forecast_time:
                continue

            # Met.no: Search for 7Timer forecast time
            if metno_index == -1:
                for datapoint in self._weather_data_metno:
                    metno_index += 1
                    if forecast_time == datetime.strptime(datapoint.get("time"), "%Y-%m-%dT%H:%M:%SZ"):
                        break
                _LOGGER.debug("Met.no start index: %s", str(metno_index))

            # Hour of day needs to be in local time
            hour_of_day = (forecast_time.hour + utc_to_local_diff) % 24

            cloudcover = row["cloudcover"]
            seeing = row["seeing"]
            transparency = row["transparency"]

            item = {
                "init": init_ts,
                "timepoint": row["timepoint"],
                "timestamp": forecast_time,
                "hour": hour_of_day,
                "seeing": seeing,
                "transparency": transparency,
                "lifted_index": row["lifted_index"],
            }
            # Met.no
            if (
                datetime.strptime(
                    self._weather_data_metno[metno_index + cnt].get("time"),
                    "%Y-%m-%dT%H:%M:%SZ",
                )
                == forecast_time
            ):
                # _LOGGER.debug("Met.no Cloud Area Fraction timestamp match: %s", str(forecast_time))
                # Continue hourly
                for i in range(0, 3):
                    details = (
                        self._weather_data_metno[metno_index + cnt + i]
                        .get("data", {})
                        .get("instant", {})
                        .get("details", {})
                    )
                    # Overwrite cloudcover
                    if details.get("cloud_area_fraction") is None:
                        break
                    item["cloudcover"] = int(details.get("cloud_area_fraction", -1) / 12.5 + 1)

                    item["cloud_area_fraction"] = details.get("cloud_area_fraction", -1)
                    item["cloud_area_fraction_high"] = details.get("cloud_area_fraction_high", -1)
                    item["cloud_area_fraction_low"] = details.get("cloud_area_fraction_low", -1)
                    item["cloud_area_fraction_medium"] = details.get("cloud_area_fraction_medium", -1)
                    item["fog_area_fraction"] = details.get("fog_area_fraction", -1)

                    item["condition_percentage"] = await self.calc_condition_percentage(
                        item["cloud_area_fraction"] / 12.5 + 1,
                        row["seeing"],
                        row["transparency"],
                    )
                    item["rh2m"] = details.get("relative_humidity", -1)
                    item["wind_speed"] = details.get("wind_speed", -1)
                    item["wind_from_direction"] = details.get("wind_from_direction", -1)
                    item["temp2m"] = details.get("air_temperature", -1)
                    item["dewpoint2m"] = details.get("dew_point_temperature", -1)
                    if self._weather_data_metno[metno_index + cnt + i].get("data", {}).get("next_1_hours", {}) == {}:
                        # No more hourly data
                        break
                    if self._weather_data_metno[metno_index + cnt + i].get("data", {}).get("next_6_hours", {}) == {}:
                        # No more 6-hourly data
                        break
                    item["weather"] = (
                        self._weather_data_metno[metno_index + cnt + i]
                        .get("data", {})
                        .get("next_1_hours", {})
                        .get("summary", {})
                        .get("symbol_code", "")
                    )
                    item["weather6"] = (
                        self._weather_data_metno[metno_index + cnt + i]
                        .get("data", {})
                        .get("next_6_hours", {})
                        .get("summary", {})
                        .get("symbol_code", "")
                    )
                    item["precipitation_amount"] = (
                        self._weather_data_metno[metno_index + cnt + i]
                        .get("data", {})
                        .get("next_1_hours", {})
                        .get("details", {})
                        .get("precipitation_amount", "")
                    )

                    items.append(ForecastData(item))

                    item["timepoint"] = item["timepoint"] + 1
                    item["timestamp"] = item["timestamp"] + timedelta(hours=1)
                    item["hour"] = item["hour"] + 1
            else:
                break

            # Limit number of Hours
            cnt += 3
            if cnt >= hours_to_show:
                break

        self._forecast_data = items

        return items

    async def _get_deepsky_forecast(self):
        """Return Deepsky Forecast data"""

        cnv = ConversionFunctions()
        items = []

        if self._forecast_data == None:
            await self._get_forecast_data(FORECAST_TYPE_HOURLY, 72)
        now = datetime.utcnow()

        # Anchor timestamp
        init_ts = await cnv.anchor_timestamp(self._weather_data_seventimer_init)

        utc_to_local_diff = self._astro_routines.utc_to_local_diff()

        # Create forecast
        forecast_dayname = ""
        start_forecast_hour = 0
        start_weather = ""
        interval_points = []

        sun_next_setting = await self._astro_routines.sun_next_setting()
        sun_next_rising = await self._astro_routines.sun_next_rising()

        _LOGGER.debug(
            "sun_next_setting: %s, sun_next_rising: %s",
            str(sun_next_setting),
            str(sun_next_rising),
        )
        for row in self._forecast_data:
            # Hour of day needs to be in local time
            hour_of_day = (row.timestamp.hour + utc_to_local_diff) % 24

            # Skip daytime, we're only interested in the forecasts at
            # darkness.
            if hour_of_day < sun_next_setting.hour and hour_of_day > sun_next_rising.hour:
                start_forecast_hour = 0
                start_weather = ""
                interval_points = []
                continue
            # cloudcover = row.cloudcover
            seeing = row.seeing
            transparency = row.transparency
            cloud_area_fraction = 0

            # Met.no
            cloud_area_fraction = row.cloud_area_fraction_percentage / 12.5 + 1

            if len(interval_points) == 0:
                forecast_dayname = row.timestamp.strftime("%A")
                start_forecast_hour = hour_of_day
                start_weather = row.weather6

            # Calculate Condition
            interval_points.append(await self.calc_condition_percentage(cloud_area_fraction, seeing, transparency))

            if hour_of_day == sun_next_rising.hour:
                item = {
                    "init": init_ts,
                    "dayname": forecast_dayname,
                    "hour": start_forecast_hour,
                    "nightly_conditions": interval_points,
                    "weather": start_weather,
                }
                items.append(NightlyConditionsData(item))

                conditions_numeric = ""
                for condition in interval_points:
                    conditions_numeric += str(condition) + ", "
                _LOGGER.debug(
                    "Nightly conditions day: %s, start hour: %s, nightly conditions: %s, weather: %s, conditions numeric: %s",
                    str(forecast_dayname),
                    str(start_forecast_hour),
                    str(len(interval_points)),
                    str(start_weather),
                    conditions_numeric,
                )

            # Forecast for two nights only
            if len(items) == 2:
                break

        return items

    async def calc_condition_percentage(self, cloudcover, seeing, transparency):
        """Return condition based on cloud cover, seeing and transparency"""
        # Possible Values:
        #   Clouds: 1-9
        #   Seeing: 1-8
        #   Transparency: 1-8
        condition = int(
            100
            - (
                self._cloudcover_weight * cloudcover
                + self._seeing_weight * seeing
                + self._transparency_weight * transparency
                - self._cloudcover_weight
                - self._seeing_weight
                - self._transparency_weight
            )
            * 100
            / (
                self._cloudcover_weight * 9
                + self._seeing_weight * 8
                + self._transparency_weight * 8
                - self._cloudcover_weight
                - self._seeing_weight
                - self._transparency_weight
            )
        )
        # _LOGGER.debug(
        #     "Calc condition cloudcover: %d(%d), seeing %d(%d), transparency: %d(%d), condition %d",
        #     cloudcover,
        #     self._cloudcover_weight,
        #     seeing,
        #     self._seeing_weight,
        #     transparency,
        #     self._transparency_weight,
        #     condition,
        # )
        return condition

    async def calc_dewpoint2m(self, rh2m, temp2m):
        """Calculate 2m Dew Point."""
        # α(T,RH) = ln(RH/100) + aT/(b+T)
        # Ts = (b × α(T,RH)) / (a - α(T,RH))
        alpha = float(Decimal(str(rh2m / 100)).ln()) + MAGNUS_COEFFICIENT_A * temp2m / (MAGNUS_COEFFICIENT_B + temp2m)
        dewpoint = (MAGNUS_COEFFICIENT_B * alpha) / (MAGNUS_COEFFICIENT_A - alpha)

        return dewpoint

    async def _get_deepsky_objects(self):
        """Return Deepsky Objects for today"""

        items = []

        await self.retrieve_data_uptonight()

        # Create list of deep sky objects
        if self._weather_data_uptonight is not None:
            dso_target_name = self._weather_data_uptonight.get("target name", {})
            dso_type = self._weather_data_uptonight.get("type", {})
            dso_constellation = self._weather_data_uptonight.get("constellation", {})
            dso_size = self._weather_data_uptonight.get("size", {})
            dso_foto = self._weather_data_uptonight.get("foto", {})

            for row in range(len(dso_target_name)):
                item = {
                    "target_name": dso_target_name.get(str(row), ""),
                    "type": dso_type.get(str(row), ""),
                    "constellation": dso_constellation.get(str(row), ""),
                    "size": dso_size.get(str(row), ""),
                    "foto": dso_foto.get(str(row), ""),
                }
                items.append(DSOUpTonight(item))

                # _LOGGER.debug(
                #     "DSO: %s, type: %s, constellation: %s, size: %s, foto: %s",
                #     str(item["target_name"]),
                #     str(item["type"]),
                #     str(item["constellation"]),
                #     str(item["size"]),
                #     str(item["foto"]),
                # )

            return items
        return None

    async def retrieve_data_seventimer(self):
        """Retrieves current data from 7timer"""

        if ((datetime.now() - self._weather_data_seventimer_timestamp).total_seconds()) > DEFAULT_CACHE_TIMEOUT:
            self._weather_data_seventimer_timestamp = datetime.now()
            _LOGGER.debug("Updating data from 7Timer")

            astro_dataseries = None
            # civil_dataseries = None

            # Testing
            if self.test_mode:
                with open("astro.json") as json_file:
                    astro_dataseries_json = json.load(json_file)
                    astro_dataseries = astro_dataseries_json.get("dataseries", {})
                    json_data_astro = {"init": astro_dataseries_json.get("init")}
                # with open("civil.json") as json_file:
                #     civil_dataseries = json.load(json_file).get("dataseries", {})
            else:
                json_data_astro = await self.async_request_seventimer("astro", "get")
                astro_dataseries = json_data_astro.get("dataseries", {})

            self._weather_data_seventimer = astro_dataseries
            self._weather_data_seventimer_init = json_data_astro.get("init")
        else:
            _LOGGER.debug("Using cached data for 7Timer")

    async def async_request_seventimer(self, product="astro", method="get") -> dict:
        """Make a request against the 7timer API."""

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            session = ClientSession(
                timeout=ClientTimeout(total=DEFAULT_TIMEOUT),
            )

        # BASE_URL_SEVENTIMER = "https://www.7timer.info/bin/api.pl?lon=XX.XX&lat=YY.YY&product=astro&output=json"
        # STIMER_OUTPUT = "json"
        url = (
            str(f"{BASE_URL_SEVENTIMER}")
            + "?lon="
            + str("%.1f" % round(self._longitude, 2))
            + "&lat="
            + str("%.1f" % round(self._latitude, 2))
            + "&product="
            + str(product)
            + "&output="
            + STIMER_OUTPUT
        )
        try:
            _LOGGER.debug(f"Query url: {url}")
            async with session.request(method, url) as resp:
                resp.raise_for_status()
                plain = str(await resp.text()).replace("\n", " ")
                data = json.loads(plain)

                if self.dump_json:
                    json_string = json.dumps(data)
                    with open(product + ".json", "w") as outfile:
                        outfile.write(json_string)

                return data
        except asyncio.TimeoutError as tex:
            raise RequestError(f"Request to endpoint timed out: {tex}") from None
        except ClientError as err:
            raise RequestError(f"Error requesting data: {err}") from None

        finally:
            if not use_running_session:
                await session.close()

    async def retrieve_data_metno(self):
        """Retrieves current data from met"""

        if ((datetime.now() - self._weather_data_metno_timestamp).total_seconds()) > DEFAULT_CACHE_TIMEOUT:
            self._weather_data_metno_timestamp = datetime.now()
            _LOGGER.debug("Updating data from Met.no")

            dataseries = None

            # Testing
            if self.test_mode:
                with open("met.json") as json_file:
                    dataseries = json.load(json_file).get("properties", {}).get("timeseries", [])
            else:
                json_data_metno = await self.async_request_met("met", "get")

                dataseries = json_data_metno.get("properties", {}).get("timeseries", [])

            self._weather_data_metno = dataseries
            self._weather_data_metno_init = dataseries[0].get("time", None)
        else:
            _LOGGER.debug("Using cached data for Met.no")

    async def async_request_met(self, product="met", method="get") -> dict:
        """Make a request against the 7timer API."""

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            session = ClientSession(
                timeout=ClientTimeout(total=DEFAULT_TIMEOUT),
            )

        # BASE_URL_MET = "https://api.met.no/weatherapi/locationforecast/2.0/complete?altitude=XX&lat=XX.XX&lon=XX.XX"
        url = (
            str(f"{BASE_URL_MET}")
            + "?lon="
            + str("%.1f" % round(self._longitude, 2))
            + "&lat="
            + str("%.1f" % round(self._latitude, 2))
            + "&altitude="
            + str(self._elevation)
        )
        try:
            _LOGGER.debug(f"Query url: {url}")
            async with session.request(method, url) as resp:
                resp.raise_for_status()
                # plain = str(await resp.text()).replace("\n", " ")
                # data = json.loads(plain)
                data = await resp.json()

                if self.dump_json:
                    json_string = json.dumps(data)
                    with open(product + ".json", "w") as outfile:
                        outfile.write(json_string)

                return data
        except asyncio.TimeoutError as tex:
            raise RequestError(f"Request to endpoint timed out: {tex}") from None
        except ClientError as err:
            raise RequestError(f"Error requesting data: {err}") from None

        finally:
            if not use_running_session:
                await session.close()

    async def retrieve_data_uptonight(self):
        """Retrieves current data from uptonight"""

        # TODO:
        # Add path to config
        # Implement Caching

        if ((datetime.now() - self._data_uptonight_timestamp).total_seconds()) > DEFAULT_CACHE_TIMEOUT:
            self._data_uptonight_timestamp = datetime.now()
            _LOGGER.debug("Updating data from uptonight")

            dataseries = None

            # Testing
            # print(os.getcwd())
            # print(os.listdir("/config/www/"))
            if os.path.isfile(self._uptonight_path + "/uptonight-report.json"):
                _LOGGER.debug(f"Uptonight report found")
                with open(self._uptonight_path + "/uptonight-report.json") as json_file:
                    dataseries = json.load(json_file)
                    _LOGGER.debug(f"Uptonight imported")
            else:
                _LOGGER.debug(f"uptonight-report.json not found. Current path: {self._uptonight_path}/uptonight-report.json")

            self._weather_data_uptonight = dataseries
        else:
            _LOGGER.debug("Using cached data for uptonight")


"""Constant Definitions for AstroWeather."""

BASE_URL_SEVENTIMER = "https://www.7timer.info/bin/api.pl"
BASE_URL_MET = "https://api.met.no/weatherapi/locationforecast/2.0/complete"
STIMER_OUTPUT = "json"

DEFAULT_TIMEOUT = 10
DEFAULT_CACHE_TIMEOUT = 1770
DEFAULT_ELEVATION = 0
DEFAULT_TIMEZONE = "Etc/UTC"
DEFAULT_CONDITION_CLOUDCOVER_WEIGHT = 3
DEFAULT_CONDITION_SEEING_WEIGHT = 2
DEFAULT_CONDITION_TRANSPARENCY_WEIGHT = 1

CIVIL_TWILIGHT = 0
CIVIL_DUSK_DAWN = -6
NAUTICAL_TWILIGHT = -6
NAUTICAL_DUSK_DAWN = -12
ASTRONOMICAL_TWILIGHT = -12
ASTRONOMICAL_DUSK_DAWN = -18

MAGNUS_COEFFICIENT_A = 17.625
MAGNUS_COEFFICIENT_B = 243.04

DEEP_SKY_THRESHOLD = 75
HOME_LATITUDE = 0.0
HOME_LONGITUDE = 0.0

CLOUDCOVER_PLAIN = [
    "0 to 6%",
    "6 to 19%",
    "19 to 31%",
    "31 to 44%",
    "44 to 56%",
    "56 to 69%",
    "69 to 81%",
    "81 to 94%",
    "94 to 100%",
]

SEEING_PLAIN = [
    'Below 0.5"',
    '0.5 to 0.75"',
    '0.75 to 1"',
    '1 to 1.25"',
    '1.25 to 1.5"',
    '1.5 to 2"',
    '2 to 2.5"',
    'Over 2.5"',
]

TRANSPARENCY_PLAIN = [
    "Below 0.3 mag",
    "0.3 to 0.4 mag",
    "0.4 to 0.5 mag",
    "0.5 to 0.6 mag",
    "0.6 to 0.7 mag",
    "0.7 to 0.85 mag",
    "0.85 to 1 mag",
    "Over 1 mag",
]

LIFTED_INDEX_PLAIN = [
    "Below -7, very unstable",
    "-7 to -5, very unstable",
    "-5 to -3, unstable",
    "-3 to 0, slightly unstable",
    "0 to 4, stable",
    "4 to 8, stable",
    "8 to 11, very stable",
    "Over 11, very stable",
]

WIND10M_DIRECTON = ["N", "NE", "E", "SE", "S", "SW", "W", "NW"]

CONDITION_PLAIN = ["excellent", "good", "fair", "poor", "bad"]
# CONDITION = ["↑", "↗", "→", "↘", "↓"]
# CONDITION = ["😁", "😀", "😐", "😕", "😞"]
CONDITION = ["█", "▆", "▄", "▂", "▁"]

FORECAST_TYPE_DAILY = "daily"
FORECAST_TYPE_HOURLY = "hourly"

FORECAST_TYPES = [
    FORECAST_TYPE_DAILY,
    FORECAST_TYPE_HOURLY,
]
